/*
 
 Usage: cf-compare [options] query_set_file [base_set_file]

 Created by: Alejandro Gutierrez
 Created Date: 11/05/2007
 Last Modified Date: 10/28/2008
 University of South Florida

 Copyright (C) 2008  Alejandro Gutierrez Munoz

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <iomanip>
#include <cmath>


using namespace std;

void CompareCurrentFlowsDatasets(const string&,const string&, const int&, const double&, const bool&);
void exit_with_help();
void parse_command_line(int,char**, string&, string&,int&, double&, bool&);



int main(int argc, char* argv[]) {

	//Global variables
	string queryfile;
	string basefile;
	int topN;
	double threshold;
	bool ProduceClassCounts;

	parse_command_line(argc,argv,queryfile,basefile,topN,threshold,ProduceClassCounts);

	// Pass the query .cfv file and the base .cfv file.
	CompareCurrentFlowsDatasets(queryfile,basefile,topN,threshold,ProduceClassCounts);

	
	return 0;
}


void exit_with_help() {
	
	cout <<	"Usage: cf-compare [options] query_set_file [base_set_file]" << endl;
	cout << "If base set is not provided, it compares the query set to itself." << endl;
	cout << "options:" << endl;
	cout << "-t (0..1): Match value threshold" << endl;
	cout << "-n (1..n): Top n best matches" << endl;  
	cout << "-c output class counts" << endl;  
	
	exit(1);
}

void parse_command_line(int argc,char* argv[], string &queryFile, string &baseFile, int &topN, double &threshold, bool &ProduceClassCounts) {

	int i;

	// default values
	topN = -1;
	threshold = 0.0;
	ProduceClassCounts = false;
	

	// parse options
	for(i=1;i<argc;i++)
	{
		if(argv[i][0] != '-') break;
		if(++i>=argc)
			exit_with_help();
		switch(argv[i-1][1])
		{
			case 'n':
				topN = atoi(argv[i]);
				if (topN < 1)
					exit_with_help();
				break;
			case 't':
				threshold = atof(argv[i]);
				if (threshold < 0.0 || threshold > 1.0 )
					exit_with_help();
				break;
			case 'c':
				ProduceClassCounts = true;
				i--;
				break;
			default:
				cerr << "unknown option" << endl;
				exit_with_help();
		}
	}

	// determine filenames

	if(i>=argc)
		exit_with_help();

	queryFile =  argv[i];

	// if there is no base set file, use same query file
	if(i<argc-1)
		baseFile = argv[i+1];
	else
		baseFile = queryFile;

	return;

}

void CompareCurrentFlowsDatasets(const string &queryDataset,const string &baseDataset, const int &topN, const double &threshold, const bool &ProduceClassCounts) {

	// First, load the query dataset
	ifstream queryFile;
	queryFile.open(queryDataset.c_str());
	bool QueryBaseEqual;

	QueryBaseEqual = (queryDataset.compare(baseDataset)) ? false : true;	

	// check if we were able to open the query file
	if (queryFile.is_open()) {

		// Second, load the base dataset
		ifstream baseFile;
		baseFile.open(baseDataset.c_str());

		// check if we were able to open the base file
		if (baseFile.is_open()) {

		
			// Now that both files are opened, proceed to load them in a data structure
			int degree;
			double value;
			vector< pair< string,string > > queryGraphs, baseGraphs;
			vector< vector< double > > GeodesicCurrents(2);
			vector< vector< vector< double > > > queryCurrents, baseCurrents;

			// start reading query file
			string line, graph_name, graph_class, geodesicStr;
			istringstream buffer;

			cout << "Loading Query Dataset..." << endl;

			while (!queryFile.eof())
			{
				getline(queryFile,line);
				buffer.str(line);

				buffer >> graph_name;

				// verify that next line contains data
				if ( !graph_name.empty() ) {
				
					buffer >> graph_class;

					queryGraphs.push_back(pair< string,string >(graph_name,graph_class)); 
			
					// read all the currents for this graph
					do {

						geodesicStr.clear();
						buffer >> geodesicStr;
				
						if (!geodesicStr.empty()) { 

							degree = atoi(geodesicStr.substr(2,1).c_str());

							// increase the size of the shortest geodesic current vector to hold new degrees
							if ( (int) GeodesicCurrents[0].size() < degree + 1 ) {
								GeodesicCurrents[0].resize(degree + 1);
							}

							// increase the size of the longest geodesic current vector to hold new degrees
							if ( (int) GeodesicCurrents[1].size() < degree + 1 ) {
								GeodesicCurrents[1].resize(degree + 1);
							}

							value = atof(geodesicStr.substr(4,geodesicStr.length()-4).c_str());
						
							if (geodesicStr.at(0) == 'S' || geodesicStr.at(0) == 's' ) {
								GeodesicCurrents[0][degree] = value ;  
							}

							if (geodesicStr.at(0) == 'L' || geodesicStr.at(0) == 'l' ) {
								GeodesicCurrents[1][degree] = value ; 
							}
						}

					} while (!geodesicStr.empty());

					queryCurrents.push_back(GeodesicCurrents); 

				}

		
				GeodesicCurrents[0].clear(); // clear shortest geodesic currents
				GeodesicCurrents[1].clear(); // clear longest geodesic currents

				buffer.clear();
				graph_name.clear();
				graph_class.clear();
				geodesicStr.clear();

				
			}

			cout << "Loading Base Dataset..." << endl;

			map< string,int > base_classes;
			map< string,double > max_MV;
					
			// Now load the base file
			while (!baseFile.eof())
			{
				getline(baseFile,line);
				buffer.str(line);

				buffer >> graph_name;

				// verify that next line contains data
				if ( !graph_name.empty() ) {

					buffer >> graph_class;

					if (ProduceClassCounts) {
					
						// gather the different classes present in the base file
						if ( !base_classes[graph_class] ) {
							base_classes.insert(pair< string,int > (graph_class,0)); 
							max_MV.insert(pair< string,double > (graph_class,0.0));
						}

					}

					baseGraphs.push_back(pair< string,string >(graph_name,graph_class)); 
			
					// read all the currents for this graph
					do {

						geodesicStr.clear();
						buffer >> geodesicStr;
				
						if (!geodesicStr.empty()) { 
						
							degree = atoi(geodesicStr.substr(2,1).c_str());

							// increase the size of the shortest geodesic current vector to hold new degrees
							if ( (int) GeodesicCurrents[0].size() < degree + 1 ) {
								GeodesicCurrents[0].resize(degree + 1);
							}

							// increase the size of the longest geodesic current vector to hold new degrees
							if ( (int) GeodesicCurrents[1].size() < degree + 1 ) {
								GeodesicCurrents[1].resize(degree + 1);
							}

							value = atof(geodesicStr.substr(4,geodesicStr.length()-4).c_str());
						
							if (geodesicStr.at(0) == 'S' || geodesicStr.at(0) == 's' ) {
								GeodesicCurrents[0][degree] = value ;  
							}

							if (geodesicStr.at(0) == 'L' || geodesicStr.at(0) == 'l' ) {
								GeodesicCurrents[1][degree] = value ; 
							}

						}

					} while (!geodesicStr.empty());

					baseCurrents.push_back(GeodesicCurrents); 
				}

			
				GeodesicCurrents[0].clear(); // clear shortest geodesic currents
				GeodesicCurrents[1].clear(); // clear longest geodesic currents

				buffer.clear();
				graph_name.clear();
				graph_class.clear();
				geodesicStr.clear();

			}


			// Now that both files were loaded in a data structure, we can begin the comparison

			// create the output file
			ofstream ResultsFile, ClassCountsFile;
			string filename;
			ostringstream str_query_instance, str_counts, str_output;

			double g1_shortest, g2_shortest, g1_longest, g2_longest, d1, d2, mv;
			int total_degrees, topNresults;
			long total_comparisons, progress;
			multimap< double,pair< string,string > >::reverse_iterator rit;
			multimap< double,pair< string,string > > topN_matches;
			map< string,int > ::iterator it;
			map< string,double > ::iterator itd;
			vector< map< string,int > > topN_counts;
			
			
			// name the results file
			filename.clear();
			filename = queryDataset.c_str() ;
			filename.append(".results");   
			ResultsFile.open(filename.c_str());

			if (ProduceClassCounts) {

				filename.clear();
				filename = queryDataset.c_str() ;
				filename.append("_ClassCounts.csv");   
				ClassCountsFile.open(filename.c_str());

				//write header of the class counts file
				ClassCountsFile << "graph,graph_class,top";

				// a column for each of the available classes plus the highest match value for each class
				for( it=base_classes.begin() ; it != base_classes.end(); it++ ) {
						ClassCountsFile << "," << (*it).first << ",MaxMV_" << (*it).first ;
				}

				ClassCountsFile <<  "\n";
			
				//we need to be aware of the labels for classification purposes
				topN_counts.resize(topN+1);

			}

			cout << "Comparing Datasets..." << endl;

			total_comparisons = (int) queryGraphs.size() * (int) baseGraphs.size();
			progress = 0;

			// for each of the graphs in the query dataset
			for (int q=0; q < (int) queryGraphs.size(); q++ ) {

				topN_matches.clear();

				str_query_instance.str("");
				str_query_instance << queryGraphs[q].first << "," << queryGraphs[q].second;

				// compare to each of the graphs in the base dataset
				for (int b=0; b < (int) baseGraphs.size(); b++ ) {

					progress++;

					//clean the values
					g1_shortest = 0.0;
					g2_shortest = 0.0;
					g1_longest = 0.0;
					g2_longest = 0.0;
					d1 = 0.0;
					d2 = 0.0;
					mv = 0.0;
					
				
					// use the graph with higher degree number
					total_degrees = max( (int) queryCurrents[q][0].size(), (int) baseCurrents[b][0].size()); 
								 
					for (int d=0; d < total_degrees ; d++ ) {


						// in case one of the graphs has fewer degrees than the other, default value to 0.0
						g1_shortest = (d > (int) queryCurrents[q][0].size() - 1 ) ? 0.0 : queryCurrents[q][0][d];
						g2_shortest = (d > (int) baseCurrents[b][0].size() - 1 ) ? 0.0 : baseCurrents[b][0][d];

						g1_longest = (d > (int) queryCurrents[q][1].size() - 1 ) ? 0.0 : queryCurrents[q][1][d];
						g2_longest = (d > (int) baseCurrents[b][1].size() - 1 ) ? 0.0 : baseCurrents[b][1][d];
					
						if (g1_shortest == g2_shortest){
							d1 += 0.0;  // No difference
						}
						else {
							d1 +=  fabs(g1_shortest - g2_shortest) / (g1_shortest + g2_shortest); // Partial Difference
						}

						if (g1_longest == g2_longest){
							d2 += 0.0;  // No difference
						}
						else {
							d2 +=  fabs(g1_longest - g2_longest) / (g1_longest + g2_longest); // Partial Difference
						}
					
					}

					mv = 1.0 - ( (d1 + d2) / (total_degrees * 2.0) );   //  total_degrees * 2 is to account for d shortest, and d longest geodesics

					if (mv >= threshold)  {
	
						// we are using here a STL multimap, which is an associative
						// container stored as a tree ordered by the first element.
						// insertion is logarithmic in size
						topN_matches.insert(pair< double, pair< string,string > > (mv,baseGraphs[b]) );
					}
					
				}

				// select only the topN results (if passed as parameter)
				topNresults = (topN == -1)? (int) topN_matches.size() : topN;

				// return the results using a reverse iterator on the multimap
				str_output.str("");
				
				//Initialize topN counts for all Base Classes
				if (ProduceClassCounts) {
					topN_counts[0] = base_classes;
				}
				
				
				for( rit=topN_matches.rbegin() ; rit != topN_matches.rend() && topNresults; rit++ ) {

					// check if the current graph is being compared to itself, if not, then use the result
					if ( !(QueryBaseEqual && queryGraphs[q].first.compare((*rit).second.first) == 0)) {
							
						// iterator object:
						// rit.first = match value
						// rit.second = graph_pair
						// graph_pair.first = graph name
						// grap_pair.second = graph class
						
						str_output << " " << (*rit).second.first << ":" << (*rit).second.second << ":" << setiosflags(ios::fixed) << setprecision(5) << (*rit).first;

						topNresults--;  // if not, then use the result
						
					
						// we now count the classes for classification purposes
						if (ProduceClassCounts) {

							// propagate TopN-1 results
							topN_counts[topN - topNresults] = topN_counts[topN - topNresults - 1];

							// increase count for g2 class
							topN_counts[topN - topNresults][(*rit).second.second]++;

							//store the first match value encountered for each class
							if (max_MV[(*rit).second.second]==0.0)
								max_MV[(*rit).second.second] = (*rit).first;

						}
						
					}
										
				}


				//write class counts to file
				if (ProduceClassCounts) {
	
					for (int i=1; i < topN_counts.size(); i++) {
						
						str_counts.str("");
						str_counts << "," << i;

						for( it=topN_counts[i].begin() ; it != topN_counts[i].end(); it++ ) {
							
							// write the count per class as well as the maximum Match Value for each class at each TopN
							str_counts <<  "," << (*it).second << "," << setiosflags(ios::fixed) << setprecision(5) << max_MV[(*it).first] * (double)(bool)(*it).second;

							// reset the counts for next graph
							topN_counts[i][(*it).first] = 0;
						}

						ClassCountsFile << str_query_instance.str() << str_counts.str() << "\n";
					}

					// reset the counts for next graph
					for( itd= max_MV.begin(); itd != max_MV.end(); itd++ ) {
						max_MV[(*itd).first] = 0.0;
					}
				}

				
				ResultsFile << str_query_instance.str() << str_output.str() << "\n";

				// report the progress
				cout << "\b\b\b\b\b\b\b\b\b\b\b\\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\\b\b\b\b\b\b\b\b\b"  // this clears the line
					<< setiosflags(ios::fixed) << setprecision(2) << (progress / (double) total_comparisons)*100.0 << "%       "; 
			
			}



			if (ProduceClassCounts) {
				ClassCountsFile.close();
			}

			ResultsFile.close();

		}
		else {
			 cerr << "Unable to open file " << baseDataset; 
		}

	}
	else {
		 cerr << "Unable to open file " << queryDataset; 
	}
	
	cout << endl << "Done." << endl;
	return;
}
