/*

 usage: Current_Flow_Vectors <inputfile.gds>
 Created by: Alejandro Gutierrez
 Created Date: 08/05/2007
 Last Modified Date: 09/25/2008
 University of South Florida


 Copyright (C) 2008  Alejandro Gutierrez Munoz

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <fstream>
#include <sstream>
#include <map>
#include <iomanip>
#include <algorithm>
#include <cmath>
#include <vector>

#include "graph.h"

using namespace std;

void CreateCurrentFlowsDataset(const string&);
int LineCount(const string&);

int main(int argc, char* argv[])
{
	
	string filename;

	if (argc != 2) {
		cerr << "usage: cf-vectors <inputfile.gds>" << endl;
	}
	else {
		filename = argv[1];
		
		// Pass a .gds (graph dataset file) and produce a .cfv (current flows vectors)
		CreateCurrentFlowsDataset(filename);
	}
	
	return 0;
}

/*
This Function creates a file that stores the current flows
for all of the shortest and longest geodesics

The format of the file to be produced is:

GraphName Class [S:d:value]* [L:d:value]*

S:d - shortest geodesic of degree d
L:d - longest geodesic of degree d

e.g.

654454 CA S:1:0.00154 S:2:1.0000 S:3:0.0285 L:1:0.55654 L:2:0.89875 L:3:0.002565
52124  CI S:1:0.40154 S:2:0.0665 L:1:1.0000 L:2:0.55655

The above lines represent two graphs, the first one 654454 belongs to the class CA
and has shortest and longest geodesics of degree 1, 2, and 3. The second graph
52124 belongs to the class CI. This one has only shortest and longest geodesics of
degree 1 and 2.


*/
void CreateCurrentFlowsDataset(const string &filename) {

	cout << flush <<  "Processing File: " << filename <<  endl;

	// First load the .gds file (graph dataset file)
	ifstream gdsFile;
	gdsFile.open(filename.c_str());

	// check if we were able to open the file
	if (gdsFile.is_open()) {

		// count how many lines are in the file, this is to track progress
		int file_lines_no;
		file_lines_no = LineCount(filename);

		int lineno = 0, number_of_graphs = 0;
		string line, graph_string;			// temp variables to load the graph definition

		graph::graphCurrents gTempCurrents; // data structure that holds the current flows

		vector< graph > graph_dataset;   // vector of graphs
		
		map< string,double > label_counts;
		map< string,double > ::iterator it;

		multimap< double,string > sorted_label_counts;
		multimap< double,string > ::reverse_iterator rit;

		double max_resistance = 0.0;
		double min_resistance = 9.99e99;

		double resistance_increment;

		// create the output file
		ofstream OutputFile;
		string output_filename;

		output_filename = filename.c_str() ;
		output_filename.append(".cfv");   // the name of the output file is the same plus .cfv
		OutputFile.open(output_filename.c_str());

		while (!gdsFile.eof())
		{
		  lineno++;
		  getline(gdsFile,line);

		  // start reading a graph
		  if (!line.compare(0,5,"BEGIN")) {

				number_of_graphs++;

				// clear current string
				graph_string.clear(); 
				graph_string.append(line);
				graph_string.append("\n");
				
				do {
					lineno++;
					getline(gdsFile,line);
					graph_string.append(line);
					graph_string.append("\n");

				} while (line.compare(0,3,"END"));
		
			  // a graph definition is now loaded, proceed to create graph object
			  try {
		
				  // store graph in graph dataset vector
				  graph_dataset.push_back(graph::graph(graph_string,label_counts,max_resistance,min_resistance));

			  }
			  catch (int) {
				cerr << "Error reading file line: " << lineno << endl; 
			  }

			  
		  } // end if
			 
		} // end while


		// determine the label's resistence based on their frequency
		
		// first insert label counts into a multimap for automatic sorting on the value
		for(it=label_counts.begin() ; it != label_counts.end(); it++ )
		{
			sorted_label_counts.insert( pair< double,string > ((*it).second,(*it).first));
		}
		
		// based on the number of labels and min-max resisntances determine the increment value
		//resistance_increment = (max_resistance -  min_resistance) / (double) sorted_label_counts.size();

		// use only up to 10% of the minimum resistence
		resistance_increment = min_resistance*0.1 / (double) sorted_label_counts.size();
		//resistance_increment= 0.0; 

		int i = 1;
		// assign the values to label_counts map to store the resistance value per label
		for(rit=sorted_label_counts.rbegin() ; rit != sorted_label_counts.rend(); rit++ )
		{
			label_counts[(*rit).second] = resistance_increment * (double)i;
			i++;
		}
		
		// Now with the values for the label's resistances set, proceed to calculate the
		// currents for every graph in the dataset.


		for(i=0; i < (int) graph_dataset.size(); i++)
		{
			try {

				// calculate the currents;
				gTempCurrents = graph_dataset[i].getCurrents(label_counts);

				// store the currents in the output file
				OutputFile << graph_dataset[i].getGraphName() << "\t" << graph_dataset[i].getGraphClass();

				// for all of the available shortest geodesics
				for (int d=0; d < (int) gTempCurrents[0].size(); d++ ) {

					// omit geodesics with current flow = 0.0
					if ( gTempCurrents[0][d] != 0.0  ) {
						OutputFile << "\tS:" << d << ":" << setiosflags(ios::fixed) << setprecision(10) << gTempCurrents[0][d];
					}
				
				}

				// for all of the available longest geodesics
				for (int d=0; d < (int) gTempCurrents[1].size(); d++ ) {

					// omit geodesics with current flow = 0.0
					if ( gTempCurrents[1][d] != 0.0  ) {
						OutputFile << "\tL:" << d << ":" << setiosflags(ios::fixed) << setprecision(10) << gTempCurrents[1][d];
					}
				
				}

				OutputFile << "\n";

			
				// report the progress
				cout << "\b\b\b\b\b\b\b\b\b\b\b\\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\\b\b\b\b\b\b\b\b\b"  // this clears the line
					<< setiosflags(ios::fixed) << setprecision(2) << ((double) (i+1) / (double) number_of_graphs)*100.0 << "%\tGraph No: " << (i+1) << " "; 

			  }
			  catch (int) {
				cerr << "Error reading file line: " << lineno << endl; 
			  }
			  
		}


		gdsFile.close();
		OutputFile.close();
		graph_dataset.clear();

	 }
	 else {
		 cout << "Unable to open file"; 
	 }

	return;
}



int LineCount(const string &filename){
	
    ifstream in(filename.c_str());
	int total = count(istreambuf_iterator<char>(in),
                      istreambuf_iterator<char>(),'\n');
   
    return total;
}




